/* -*-  Mode:java; c-basic-offset:4; tab-width:4; indent-tabs-mode:nil -*- */
/*
 * ICMPTrainRecordReader.java
 *
 * Copyright (C) 2007-2009 by USC/ISI
 * $Id: ICMPTrainRecordReader.java 6596 2010-01-27 01:06:37Z yuri $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

package edu.isi.hadoop.icmptrain;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.EOFException;
import java.io.InputStream;
import java.io.DataInputStream;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.ListIterator;
import java.util.Vector; //import java.net.InetAddress;
import org.apache.commons.logging.Log;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.Seekable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.compress.CodecPool;
import org.apache.hadoop.io.compress.CompressionCodecFactory;
import org.apache.hadoop.io.compress.CompressionCodec;
import org.apache.hadoop.io.compress.SplitCompressionInputStream;
import org.apache.hadoop.io.compress.SplittableCompressionCodec;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.FileSplit;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.RecordReader;

/**
 * This class can serve as input reader (option <tt>-inputreader ICMPTrainRecordReader</tt> 
 * to hadoop streaming), for reading LANDER icmptrain data files. Due to deprecation 
 * of <tt>-inputreader</tt> the newer 
 * <tt>-inputformat edu.isi.hadoop.icmptrain.ICMPTrainInputFormat</tt> option should be used.
 * <p>
 * This reader parses and prints out contents of binary icmptrain output file, possibly compressed.
 * These are the files that we use to record data from Internet censuses and surveys.
 * <p>
 * <h3><u>CONFIGURATION</u></h3>
 * What is printed can be configured using jobconf (option <tt>-D</tt>) as follows:
 * <ul>
 * 	<li><tt>stream.recordreader.icmptrain.keys</tt> - specify fields used as keys
 * 	<li><tt>stream.recordreader.icmptrain.vals</tt> - specify fields used as values
 * </ul>
 * Fields can be:
 * <center>
 * <table border="1">
 *      <tr><th>Name</th><th>Length</th><th>Comment</th></tr>
 * 	<tr><th><tt>type</tt></th><td>1 byte</td>
 *          <td>ICMP type field</td></tr> 
 * 	<tr><th><tt>code</tt></th><td>1 byte</td>
 *          <td>ICMP code field</td></tr>
 * 	<tr><th><tt>typeandcode</tt></td><td>2 bytes</td>
 *          <td>ICMP type and code concatenated</td></tr>
 *	<tr><th><tt>typecode</tt></th><td>1 byte</td>
 *          <td>ICMP type and code fiedls bunched together (legacy)</td></tr>
 *	<tr><th><tt>ttl</tt></th><td>1 byte</td>
 *          <td>ICMP response remaining time to live</td></tr>
 *	<tr><th><tt>time</tt></th><td>4 bytes</td>
 *          <td>Seconds since the Epoch that the probe was sent</td></tr>
 *	<tr><th><tt>rtt</tt></th><td>4 bytes</td>
 *          <td>Round trip time in microseconds (if replied or zero)</td></tr>
 *      <tr><th><tt>flag</tt>s</th><td>1 byte</td>
 *          <td>Format flags</td></tr>
 *	<tr><th><tt>probeaddr</tt></th><td>4 bytes</td>
 *          <td>probed IP adddress (or zero if didn't match)</td></tr>
 *	<tr><th><tt>replyaddr</tt></th><td>4 bytes</td>
 *          <td>IP address of the responder (if any or zero)</td></tr>
 * </table>
 * </center>
 * These fields can be grouped by using underscore('_') or comma(',') as a 
 * separator.
 * <p>
 * Examples:
 * <pre>
 * 	-D stream.recordreader.icmptrain.keys=probeaddr
 * 	-D stream.recordreader.icmptrain.vals=type_code_typecode
 * 	-D stream.recordreader.icmptrain.vals=type,code,typeandcode
 * </pre>
 * As mentioned earlier that data to this reader can be a compressed stream.  
 * Few codecs split the input file into chunks while many process one whole file.
 * If a codec can handle split file, it will be done so seamlessly.  But this behavior
 * can be over-ridden by setting option mapred.input.codec.nosplitting to true
 * as follows:
 * <pre>
 * 	-D mapred.input.codec.nosplitting=true
 * </pre>
 * As said earlier that ICMP data is binary.  So whenever ICMP file is split, this
 * reader needs to align itself with an ICMP record boundary.  This reader uses
 * few heuristics for that.  The strength of this heuristic can be controlled by
 * the option <tt>stream.recordreader.icmptrain.lookaheadcount</tt> as follows:
 * <pre>
 * 	-D stream.recordreader.icmptrain.lookaheadcount=100
 * </pre>
 * In this case the reader will use 100 ICMP records to verify that its alignment
 * choice is right.
 * 
 * <h3><u>OUTPUT</u></h3> 
 * All keys and values are printed in hex (without leading 0x) in
 * fixed length format (number of hex digits being the number of bytes times
 * two). All fields are separated by TABs.
 * 
 * <h3><u>USAGE</u></h3>
 * <pre>
 *  HADOOP=$HADOOP_HOME/bin/hadoop
 *  STREAMING_JAR=$HADOOP_HOME/build/hadoop-streaming.jar
 *  INPUTREADER_JAR=$HADOOP_HOME/build/hadoop-icmptrain.jar
 *  INPUTREADER_CLASS=edu.isi.hadoop.icmptrain.ICMPTrainRecordReader ... export
 *  INPUTFORMAT_CLASS=edu.isi.hadoop.icmptrain.ICMPTrainInputFormat
 *  HADOOP_CLASSPATH=${INPUTREADER_JAR} ... $HADOOP jar $STREAMING_JAR -file
 *  $INPUTREADER_JAR -inputreader $INPUTREADER_CLASS ... ...
 *  or
 *  $INPUTREADER_JAR -inputformat $INPUTFORMAT_CLASS ... ...
 * </pre>
 */

public class ICMPTrainRecordReader implements RecordReader<Text, Text> {
    /**
     * This class provides interface for reading binary files such as produced
     * by 'icmptrain'
     */
    final static int DEFAULT_LOOKAHEAD_COUNT = 10;

    private interface ICMPTrainDataRecord {

        final static String CONF_NS = "stream.recordreader.icmptrain.";

        int toBoundary(DataInputStream in, int recordLookAheadCount) throws IOException;

        long readRecord(DataInputStream in) throws IOException, EOFException;
        
        boolean isData() throws IOException;

        boolean isValidRecord(byte records[], int index);
	
        void format(Text key, Text value);
    }
    
    private static class ICMPTrainDataRecordv2 implements ICMPTrainDataRecord {
        final static int RECORD_BYTES = 24;
        final static byte IPR_TYPE_DATA_v2 = 3;
        final static byte IPR_TYPE_TEXT_v2 = 4;
        final static String[] FIELDS = { "type", // 0
                                         "code", // 1
                                         "ttl", // 2
                                         "time", // 3
                                         "rtt", // 4
                                         "probeaddr", // 5
                                         "replyaddr", // 6
                                         "typecode",  // 7
                                         "typeandcode", //8
                                         "flags", //9
        };
        /**
         * toBoundary method reads data from the stream to find
         * the next ICMP record. This method reads (recordLookAheadCount + 1) records
         * worth of data from the stream. But this data is also needed in actual
         * processing. So alignment method uses the Java facility of mark /
         * reset so that when the call returns from alignment method, the stream
         * is exactly at the same position as before the call.
         * 
         * We use (recordLookAheadCount + 1) records worth of buffer because in the
         * first 24 bytes we try to find either of text or data suspects.  Then
         * we use recordLookAheadCount number of records for the verification. 
         */
        public int toBoundary(DataInputStream input, int recordLookAheadCount) 
            throws IOException 
        { 
            int readBound = ICMPTrainDataRecordv2.RECORD_BYTES*(recordLookAheadCount+1);
            byte rawData[] = new byte[readBound];
            input.mark(readBound);
            
            int count = 0;
            for (int offset = 0; offset < readBound; offset += ICMPTrainDataRecordv2.RECORD_BYTES) {
                if (input.read(rawData, offset, 1) != 1) {
                    LOG.warn("ICMPTrainDataRecordv2::toBoundary:");
                    LOG.warn("  EOF before all requested records could be read: " + 
                             count + " out of " + (recordLookAheadCount+1));
                    break; //eof
                }
                input.readFully(rawData, offset+1, ICMPTrainDataRecordv2.RECORD_BYTES-1);
                ++count;
            }
            LOG.info("ICMPTrainDataRecordv2::toBoundary: read " + count + " records");
            
            Vector<Integer> suspects = new Vector<Integer>();
            // try all possible byte offsets
            for (int offset = 0; offset < ICMPTrainDataRecordv2.RECORD_BYTES; ++offset) {
                if (verifyOffset(rawData, offset, count-1)) {			
                    suspects.add(offset);
                }
            }
            input.reset();
            
            if (suspects.size() == 0) {
                throw new IOException("Failed to find an ICMPTrain Record Boundary.");
            } else if (suspects.size() > 1) {
                LOG.warn("First option is chosen amongst more than one options as valid ICMP record start");
            }
	    
            return suspects.get(0);
        }//end of toBoundary method
        
        public boolean isValidRecord(byte records[], int index) {
            // Length should have value 24
            byte ipr_len = records[index + 1];
            if (ipr_len != ICMPTrainDataRecordv2.RECORD_BYTES) return false;
            
            byte ipr_type = records[index];
            switch (ipr_type) {
            case ICMPTrainDataRecordv2.IPR_TYPE_DATA_v2:
                int reserved = records[index + 4] | records[index + 5];
                if (reserved != 0) 
                    return false;
                break;
            case ICMPTrainDataRecordv2.IPR_TYPE_TEXT_v2:
                break;
            default:
                return false;
            }
            return true;
        }//end of isValidRecord method
	
	
        private boolean verifyOffset(byte records[], int offset, int recordLookAheadCount)
            throws IOException
        {
            for (int i = 0; i < recordLookAheadCount; ++i) {
                int index = offset + (ICMPTrainDataRecordv2.RECORD_BYTES * i);
                if (!isValidRecord(records, index))
                    return false;
            }
            return true;
        }//end of verifyOffset method

        public ICMPTrainDataRecordv2(JobConf j) {
            // populate the map used for field lookup and parsing
            Integer dex = 0;
            for (String s : FIELDS) {
                fieldMap_.put(s, dex);
                ++dex;
            }
            // Get the configuration
            String kfields = j.get(CONF_NS + "keys");
            if (kfields == null)
                kfields = "probeaddr"; // default
	    
            String[] kfs = kfields.split("[_,]");
            for (String s : kfs) {
                dex = fieldMap_.get(s);
                if (dex == null) {
                    LOG.error("invalid field name " + s);
                } else {
                    kfields_.add(dex);
                }
            }
	    
            String vfields = j.get(CONF_NS + "vals");
            if (vfields == null)
                vfields = "typecode"; // default
            String[] vfs = vfields.split("[_,]");
            for (String s : vfs) {
                dex = fieldMap_.get(s);
                if (dex == null) {
                    LOG.error("invalid field name " + s);
                } else {
                    vfields_.add(dex);
                }
            }
        }
	
        public long readRecord(DataInputStream in) throws IOException, EOFException {
            b_[0] = in.readByte(); // can throw EOFException, and it's OK
            // If first byte is read and the rest fails, we're in trouble
            try {
                in.readFully(b_, 1, b_.length - 1);
                ipr_time_s_ = in.readInt();
                ipr_rtt_us_ = in.readInt();
                ipr_probe_addr_ = in.readInt();
                ipr_reply_addr_ = in.readInt();
            } catch (EOFException e) {
                // EOF here, after the first byte is read means the file is
                // truncated
                throw new IOException("Input file is truncated");
            }
            return RECORD_BYTES;
        }
	
        public boolean isData() throws IOException {
            if (len() != RECORD_BYTES)
                throw new IOException("Bad type: " + type() + " length: " + len());
            switch (type()) {
            case IPR_TYPE_DATA_v2: return true;
            case IPR_TYPE_TEXT_v2: return false;
            default:
                throw new IOException("Bad type: " + type() + " length: " + len());
            }
        }

        public void format(Text key, Text val) {
            boolean is_first;
	    
            // format key
            String k = "";
            is_first = true;
            for (ListIterator<Integer> it = kfields_.listIterator(); it
                     .hasNext();) {
                String s = formatField(it.next());
                if (is_first)
                    k = s;
                else
                    k += "\t" + s; // XXX may be key fields should not be
                // separated by TAB?
                is_first = false;
            }
            key.set(k);
	    
            // format val
            String v = "";
            is_first = true;
            for (ListIterator<Integer> it = vfields_.listIterator(); it
                     .hasNext();) {
                String s = formatField(it.next());
                if (is_first)
                    v = s;
                else
                    v += "\t" + s;
                is_first = false;
            }
            val.set(v);
        }

        private String formatField(int num) {
            String x;
            switch (num) {
            case 0:
                x = toHex(reply_type(), 2);
                break;
            case 1:
                x = toHex(reply_code(), 2);
                break;
            case 2:
                x = toHex(ttl(), 2);
                break;
            case 3:
                x = toHex(time_s(), 8);
                break;
            case 4:
                x = toHex(rtt_us(), 8);
                break;
            case 5:
                x = probe_addr();
                break;
            case 6:
                x = reply_addr();
                break;
            case 7:
                // old format
                int typecode = (reply_type() << 4) | (reply_code());
                x = toHex(typecode, 2);
                break;
            case 8:
                // new format
                int typeandcode = (reply_type() << 8) | (reply_code());
                x = toHex(typeandcode, 4);
                break;
            case 9:
                x = toHex(flags(), 2);
                break;
            default:
                x = "error";
                // xxx
            }
            return x;
        }
	
        public byte type() {
            return b_[0];
        }
	
        public byte len() {
            return b_[1];
        }
	
        public byte reply_type() {
            return b_[2];
        }
	
        public byte reply_code() {
            return b_[3];
        }
        //4 and 5 are reserved (=0)
        public byte flags() {
            return b_[6];
        }
	
        public byte ttl() {
            return b_[7];
        }
	
        public int time_s() {
            return ipr_time_s_;
        }
	
        public int rtt_us() {
            return ipr_rtt_us_;
        }
	
        public String probe_addr() {
            return toHex(ipr_probe_addr_, 8);
        }
	
        public String reply_addr() {
            return toHex(ipr_reply_addr_, 8);
        }
	
        public static String toHex(int a, int maxlen) {
            String as = "00000000" + Integer.toHexString(a);
            try {
                return as.substring(as.length() - maxlen);
            } catch (java.lang.IndexOutOfBoundsException e) {
                LOG.error("can't convert " + a + " to hex");
                return "xxxxxxxx";
            }
        }

        private byte[] b_ = new byte[8];
        private int ipr_time_s_;
        private int ipr_rtt_us_;
        private int ipr_probe_addr_;
        private int ipr_reply_addr_;
	
        private LinkedList<Integer> kfields_ = new LinkedList<Integer>();
        private LinkedList<Integer> vfields_ = new LinkedList<Integer>();
        private HashMap<String, Integer> fieldMap_ = new HashMap<String, Integer>();
    }

    private static class ICMPTrainDataRecordv3 extends ICMPTrainDataRecordv2 {
        final static byte IPR_TYPE_DATA_v3 = 5;
        final static byte IPR_TYPE_TEXT_v3 = 6;
        public ICMPTrainDataRecordv3(JobConf j) { super(j); }
        public boolean isValidRecord(byte records[], int index) {
            // Length should have value 24
            byte ipr_len = records[index + 1];
            if (ipr_len != RECORD_BYTES) return false;

            byte ipr_type = records[index];
            switch (ipr_type) {
            case IPR_TYPE_DATA_v3:
                int reserved = records[index + 4] | records[index + 5];
                if (reserved != 0) 
                    return false;
                break;
            case IPR_TYPE_TEXT_v3:
                break;
            default:
                return false;
            }
            return true;
        }//end of isValidRecord method

        public boolean isData() throws IOException {
            if (len() != RECORD_BYTES)
                throw new IOException("Bad type: " + type() + " length: " + len());
            switch (type()) {
            case IPR_TYPE_DATA_v3: return true;
            case IPR_TYPE_TEXT_v2: return false;
            default:
                throw new IOException("Bad type: " + type() + " length: " + len());
            }	    
        }
    }

    private class ICMPTrainDataRecordv2_3 extends ICMPTrainDataRecordv3 {
        public ICMPTrainDataRecordv2_3(JobConf j) { super(j); }
        public boolean isValidRecord(byte records[], int index) {
            // Length should have value 24
            byte ipr_len = records[index + 1];
            if (ipr_len != RECORD_BYTES) return false;

            byte ipr_type = records[index];
            switch (ipr_type) {
            case IPR_TYPE_DATA_v2:
            case IPR_TYPE_DATA_v3:
                int reserved = records[index + 4] | records[index + 5];
                if (reserved != 0) 
                    return false;
                break;
            case IPR_TYPE_TEXT_v2:
            case IPR_TYPE_TEXT_v3:
                break;
            default:
                return false;
            }
            return true;
        }//end of isValidRecord method

        public boolean isData() throws IOException {
            if (len() != RECORD_BYTES)
                throw new IOException("Bad type: " + type() + " length: " + len());
            switch (type()) {
            case IPR_TYPE_DATA_v2:
            case IPR_TYPE_DATA_v3:
                return true;
            case IPR_TYPE_TEXT_v2:
            case IPR_TYPE_TEXT_v3:
                return false;
            default:
                throw new IOException("Bad type: " + type() + " length: " + len());
            }	    
        }
    }

    /**
     * Following class provides the facility of mark / reset.  We need
     * this facility because, when ICMPTrain constructor is called, it
     * needs to align itself to ICMP record boundary.  For this purpose
     * it needs to read many bytes from the stream.  But once a correct
     * alignment is done, all these read bytes (for alignment purposes) must
     * be written back to the stream.*/
    private static class ICMPBufferedInputStream extends BufferedInputStream {
        private boolean enableBufferSizeReporting = true;
        public ICMPBufferedInputStream(InputStream localIn, int bufferSize) {
            super(localIn, bufferSize);
        }
	
        //The following method is only meaningful when the input is
        //plain text.  In case of compressed input, position is solely
        //handeled by the specific codec class.  So for compressed input
        //the bolean enableBufferSizeReporting should be set to false
        //by the client of this class.
        public int getBufferedByteCount() {
            // return the remaining bytes in the buffer
            return (enableBufferSizeReporting==true) ? (count - pos) : 0;
        }
		
        public void setBufferSizeReporting(boolean val){
            enableBufferSizeReporting = val;
	    
        }
	
    }

    // This constructor is used in Hadoop streaming
    public ICMPTrainRecordReader(FSDataInputStream in, FileSplit split,
                                 Reporter reporter, JobConf job, FileSystem fs) 
        throws IOException, NumberFormatException {
        
        this(job, split);
    }

    // More general constructor
    public ICMPTrainRecordReader(Configuration job, FileSplit split)
        throws IOException, NumberFormatException {
	
        start_ = split.getStart();
        end_ = start_ + split.getLength();
        final Path file = split.getPath();
	
        // try dealing with compressed files
        compressionCodecs_ = new CompressionCodecFactory(job);
        codec_ = compressionCodecs_.getCodec(file);
        // open the file and seek to the start of the split
        FileSystem fs = file.getFileSystem(job);
        FSDataInputStream fileIn = fs.open(split.getPath());
        fileIn.seek(start_);
        
	
        String noSplitting = (job.get("mapred.input.codec.nosplitting")==null)?
            "false":job.get("mapred.input.codec.nosplitting");
        
        
        in = fileIn;
        if (codec_ != null) {
            if (codec_ instanceof SplittableCompressionCodec 
                && !(noSplitting.equalsIgnoreCase("true"))) {
                LOG.info("Bzip2 code being run!");
                final 
                    SplitCompressionInputStream cIn =
                    ((SplittableCompressionCodec)codec_).createInputStream(
                                                                           fileIn, 
                                                                           CodecPool.getDecompressor(codec_),
                                                                           start_, 
                                                                           end_,
                                                                           SplittableCompressionCodec.READ_MODE.BYBLOCK);
                //Codecs have changed start / end.  So we need to update them.
                start_ = cIn.getAdjustedStart();
                end_ = cIn.getAdjustedEnd();
                in = cIn;
            } else {
                LOG.info("got compression codec!");
                in = codec_.createInputStream(fileIn);
                end_ = Long.MAX_VALUE;
            }
        }
        
        String version = job.get(ICMPTrainDataRecordv2.CONF_NS + "version");
        if (version == null) {
            LOG.info("ICMPTrainDataRecordReader: v2,v3");
            rec_ = new ICMPTrainDataRecordv2_3(new JobConf(job));
        } else if (version.equals("2")) {
            rec_ = new ICMPTrainDataRecordv2(new JobConf(job));
        } else if (version.equals("3")) {
            rec_ = new ICMPTrainDataRecordv3(new JobConf(job));
        } else {
            LOG.error("Invalid version in " + ICMPTrainDataRecordv2.CONF_NS +
                      "version: " + version);
            throw new IOException("Unknown version");
        }
        String value = job.get(ICMPTrainDataRecord.CONF_NS + "lookaheadcount");
        
        int recordLookAheadCount = (value == null) 
            ? DEFAULT_LOOKAHEAD_COUNT 
            : Integer.parseInt(value);
        
        buffIn = new ICMPBufferedInputStream(in, (recordLookAheadCount*
                                                  (ICMPTrainDataRecordv2.RECORD_BYTES + 1)));
        //As commented at the definition of setBufferSizeReporting
        //method that for compressed data this buffered data size
        //statistic is meaningless because in this buffer
        //un-compressed data is placed while start_, end_ and pos_ are
        //in terms of compressed data units.
        if (codec_!=null) buffIn.setBufferSizeReporting(false);
        
        in_ = new DataInputStream(buffIn);
        int skipBytes = rec_.toBoundary(in_, recordLookAheadCount);
        LOG.info("Bytes needed to skip:" + skipBytes);
        
        /*
         * If this is not the first split (i.e. start_ != 0), then we need 
         * to throw away so many bytes that stream is positioned at the next
         * record. This is so because next() method always reads one more 
         * record beyond its allowed reading limit (of course an exception 
         * is the last split, when the stream ends and next() method can not 
         * read beyond the end of stream) If skipBytes == 0 then we need to 
         * throw away full ICMPTrainDataRecordv2.RECORD_BYTES bytes else we 
         * only need to throw away skipBytes number of bytes. 
         */
        skipBytes = (skipBytes == 0 && start_ != 0) ? 
            ICMPTrainDataRecordv2.RECORD_BYTES : skipBytes;
        if (skipBytes>0) {
            if (in_.read(new byte[1]) != 1)
                LOG.info("Failed to read first byte (EOF)?");
           else
               in_.readFully(new byte[skipBytes-1]); //this can throw an exception
        }
    }
	

    public synchronized boolean next(Text key, Text value) throws IOException {
            do {
                pos_ = getPos();
                if (pos_ > end_)
                    return false;
                try {
                    long size = rec_.readRecord(in_);
                    if (size == 0) {
                        return false;
                    }
                } catch (EOFException e) {
                    LOG.info("Caught an EOFException");
                    return false;
                }
            } while (!rec_.isData()); // skip text fields, can throw exception
            // parse the record, format key and value
            rec_.format(key, value);
            return true;
        }
    
    public Text createKey() {
        return new Text();
    }
    
    public Text createValue() {
        return new Text();
    }
    
    public float getProgress() throws IOException {
        if (start_ == end_) {
            return 0.0f;
        } else {
            pos_ = getPos();
            return Math.min(1.0f, (pos_ - start_) / (float) (end_ - start_));
        }
    }
    
    public synchronized long getPos() throws IOException {
            /* If input was plain text, we must subtract buffered data to
             * get the correct position.  In case of compressed data
             * getBufferedByteCount() should be set to return 0 because
             * otherwise we will be mixing compressed and un-compressed
             * units (getPos will be returning position in compressed
             * stream while getBufferedByteCount will be returning amount
             * of buffered un-compressed data) which is wrong.  So in case
             * of compressed input, we are dependent on the respective
             * codec class. */
            return ((Seekable) in).getPos() - buffIn.getBufferedByteCount();
        }

    public synchronized void close() throws IOException {
            in.close();
        }
	

    private long start_;
    private long end_;
    private long pos_;
    private DataInputStream in_;
    private InputStream in;
    private ICMPBufferedInputStream buffIn;
    private ICMPTrainDataRecord rec_;
    private CompressionCodecFactory compressionCodecs_;
    private CompressionCodec codec_;
    private static final Log LOG = LogFactory.getLog(ICMPTrainRecordReader.class.getName());
}
